/* View streams from a webcam with Xv */
/* Parts from an app used to test Trident drivers, and camE */
/* $Id: xvcam.c,v 1.2 2001/06/21 19:30:02 richlowe Exp $ */
/* gcc xvcam.c -o xvcam -I/usr/X11R6/include -L/usr/X11R6/lib -lX11 -lXext -lXv -lm */

/*
 * This software is part of the Enlightenment project, and was found
 * in the camE directory. The camE project is written under GPL2.
 *
 * When I found this file, it was not licensed, but the author wishes to
 * publish under GPL2 as well.
 *
 * So, all modifications done by me ( Demaine B-P) or any other haker
 * shall be done under GPL2.
 *
 * For more details about the GPL : http://www.gnu.org/
 * For more details about Enlightenment : google -> enlightenment
 * For more details about authors of camE and xvcam,
 * see the AUTHORS file in the CVS tree.
 * For more details about DEMAINE B-P : http://www.demaine.info/
 * 
 */

/* Note of DEMAINE Benoit-Pierre:
 * my modifdied version might requires to be compiled with option -lm
 * the keyboard of my laptop is french converted in DVORAK - thus 
 * the keys read will probably be wrong for _your_ computer
 *
 * The original xvcam was about 400 lines. It acquires the picture of the webcam
 * in the YUV format I didnt know at the beginning.
 * Note that this program works fine with ov511 cams, such as my
 * Terratec Terracam PRO, but not with the Terracam Home wich is CPIA ...
 * 
 * This program was written to detect an object in a pic from a webcam
 * and update servos position using the hc08 driver.
 * See the SelfTackingWebcam project in my home page http://www.demaine.info/
 */

// The pseed of the process depends on many factors. The main ones are
// _R_RES_ and _R_WIDTH_ ; the speed of processing will be
// linearly proportionally inverse of those values.
// if you use very low values, you can have a very high rate ...
// but you cannot go above the numxer of pic/sec the cam can give ( USB1)
// which is about 4-5 p/s for me. using 360:1 or 100:3 I have 3.5p/s
// on my AMD 1G4. But it is useless to try to go higher than 5p/s.
// it all depend on your CPU and your cache.
//
// The quality depends on ( in decreasing importance )
// [recommended value/minimum recommended)
// _MAX_EDGE_ERROR_ [50/20] _ACCEPTED_RATE_ [20/05]
// _ACCEPTED_RATE__ [70/40]
// _R_RES_ [200/50] _R_WIDTH_ [3/1]

#define ASPECT_RATIO 2	// should be betmeen 1 and 4
#define SIZE	8
#define _YUV_WIDTH 320
#define _YUV_HEIGHT 240
#define _SKIP_ 25
#define _R_RES_ 50	// resolution of the R-Table - hom many points per 2*Pi
#define	_R_WIDTH_ 3	// number of points per rdius // speed factor
#define _MAX_EDGE_ERROR_ 50	// tells how many times the random 
				// num generator searchs for usefull points
				// this is due to bugs of very fuzzy pics
				// or when the cam is moving too fast
#define _ACCEPTED_RATE_ 20	// miminum accuracy ratio between local R_table
				// and the object R_Table
#define _ACCEPTED_RATE__ 70
#define PRECOS_SIZE	1024
#define PI 3.141592
#define _BLACK_FILE "/tmp/black_noise.raw"

// min and max for servos
#define	_MIN_5_	0x10
#define _MAX_5_	0xF0
#define	_MIN_6_	0x30
#define _MAX_6_	0x85

#define _USB_SLEEP_TIME_	100000

int tmp_x, tmp_y, tmp_i, tmp_j;

struct plop {
	int	usb;
	unsigned char	buf[SIZE];
	unsigned char	buf_[SIZE];
	unsigned char	noise[_YUV_WIDTH*_YUV_HEIGHT*2];
	int lock;
	int oldx, oldy;
	int	tab1[_YUV_WIDTH][_YUV_HEIGHT];	// light
	int	tab1_gb[_YUV_WIDTH/2][_YUV_HEIGHT/2]; // green-blue
	int	tab1_gr[_YUV_WIDTH/2][_YUV_HEIGHT/2]; // green-red
	int	G_h[_YUV_WIDTH][_YUV_HEIGHT];	// hor edges
	int	G_v[_YUV_WIDTH][_YUV_HEIGHT];	// ver edges
	float	S[_YUV_WIDTH][_YUV_HEIGHT];	// strength
	float	P[_YUV_WIDTH][_YUV_HEIGHT];	// phase
	int e_filter;
	int filter_gb;
	int filter_gr;
	int noise_f;
	int view;
	int s_t;
	int flag;
	int R[_R_RES_][_R_WIDTH_];	// R-Table
	int R_valid[_R_RES_][_R_WIDTH_];	// R-Table
	int R_x[_R_RES_][_R_WIDTH_];	// R-Table Delta_x
	int R_y[_R_RES_][_R_WIDTH_];	// R-Table Delta_y
	float R_P[_R_RES_][_R_WIDTH_];	// phase of the edge
	int R_W[_R_RES_][_R_WIDTH_];	// width
	int mouse_x, mouse_y;
	int W[_YUV_WIDTH][_YUV_HEIGHT];	// width of an edge ...
	// it is the number of points both element of an edge and of
	// the tangant of an edge at a given point ... computed in length()
	// it is an aproximation of the "rayon de courbure"[in french]
	int search;	// tell either we must compute the HT
	int edge_factor;
	int result[_YUV_WIDTH][_YUV_HEIGHT];
	int show;	// I forgot what this is for
	int sleep;	// set if we must wait after the pic is shown
} *i2;

struct plouf {
	int R[_R_RES_];
	int R_valid[_R_RES_];
	int R_x[_R_RES_];
	int R_y[_R_RES_];
	int R_s[_R_RES_];	// strength of edge for the R-Table
} plaf;

float coss[PRECOS_SIZE];
float sinn[PRECOS_SIZE];

int film;

#define _F_ 4	// filter size -1
#define _H_ 3	// half of _F_ +1
#define _F2_ 8
#define _H2_ 6
int	filter1_h[5][5]={
	{0, 0, 0, 0, 0},
	{0, 1, -1, 0, 0},
	{0, 2, -2, 0, 0},
	{0, 1, -1, 0, 0},
	{0, 0, 0, 0, 0},
};


int	filter1_v[5][5]={
	{0, 0, 0, 0, 0},
	{0,  1,  2,  1, 0},
	{0, -1, -2, -1, 0},
	{0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0},
};

int	filter2_v[5][5]={
	{0, 0, 0, 0, 0},
	{ 1,  2,  3,  2,  1},
	{-1, -2, -3, -2, -1},
	{0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0}
};

int	filter2_h[5][5]={
	{0, -1, 1, 0, 0},
	{0, -2, 2, 0, 0},
	{0, -3, 3, 0, 0},
	{0, -2, 2, 0, 0},
	{0, -1, 1, 0, 0},
};


/**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>
#include <X11/extensions/XShm.h>

#include <linux/videodev.h>

#define GUID_I420_PLANAR 0x30323449

#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <sys/wait.h>
//#include <time.h>
#include <sched.h>
#include <png.h>

extern int XShmGetEventBase(Display *);
extern XvImage *XvShmCreateImage(Display *, XvPortID, int, char *, int, int,
                                 XShmSegmentInfo *);

int grab_fd, grab_size;
static struct video_mmap grab_buf;
int yuv_width = _YUV_WIDTH;
int yuv_height = _YUV_HEIGHT;
static unsigned char *grab_data = NULL;
static unsigned char *grab_data_tmp = NULL;

float time_diff(struct timeval *start, struct timeval *end)
{	// gives the diff time in ms.
	float r, s, us;
	s=(float)(end->tv_sec-start->tv_sec);
	us=(float)(end->tv_usec-start->tv_usec);
	r=s*1000+us/1000;
	return(r);
}

void
deinit()
{
//	struct timeval t_start, t_end;
	munmap(grab_data_tmp, grab_size);
	grab_data_tmp = NULL;
//		gettimeofday(&t_start,NULL);
	close(grab_fd);
//		gettimeofday(&t_end,NULL);
//		printf("grab: %fms\n\n",time_diff(&t_start,&t_end));
//		gettimeofday(&t_start,NULL);
	close(i2->usb);	// this is the call which takes 100% CPU when exitting
//		gettimeofday(&t_end,NULL);
//		printf("usb: %fms\n\n",time_diff(&t_start,&t_end));
//		gettimeofday(&t_start,NULL);
	free(i2);
//		gettimeofday(&t_end,NULL);
//		printf("i2: %fms\n\n",time_diff(&t_start,&t_end));
}

void
init()
{
   struct video_capability grab_cap;
   struct video_channel grab_chan;
   struct video_mbuf vid_mbuf;

   atexit(deinit);

   if ((grab_fd = open("/dev/video0", O_RDWR)) == -1) {
      fprintf(stderr, "Couldn't open /dev/video0: %s\n", strerror(errno));
      exit(-1);
   }

   if (ioctl(grab_fd, VIDIOCGCAP, &grab_cap) == -1) {
      perror("ioctl VIDIOCGCAP");
      exit(-1);
   }

   grab_chan.channel = 0;

   if (ioctl(grab_fd, VIDIOCGCHAN, &grab_chan) == -1) {
      perror("ioctl VIDOCGCHAN");
      exit(-1);
   }

   grab_chan.norm = 0;

   if (ioctl(grab_fd, VIDIOCSCHAN, &grab_chan) == -1) {
      perror("ioctl VIDIOCSCHAN");
      exit(-1);
   }

   grab_buf.format = VIDEO_PALETTE_YUV420P;
   grab_buf.frame = 0;
   grab_buf.width = yuv_width;
   grab_buf.height = yuv_height;

   ioctl(grab_fd, VIDIOCGMBUF, &vid_mbuf);

   grab_size = vid_mbuf.size;
   grab_data_tmp =
       mmap(0, grab_size, PROT_READ | PROT_WRITE, MAP_SHARED, grab_fd, 0);

   if ((grab_data_tmp == NULL) || (-1 == (int) grab_data_tmp)) {
      fprintf(stderr, "Couldn't mmap\n");
      exit(1);
   }

   /* Useless? probably. */
   setpriority(PRIO_PROCESS, 0, 20);
   nice(20);
}


void
grab_image()
{
   int i = 0;
   fd_set fds;

   FD_ZERO(&fds);
   FD_SET(0, &fds);
   FD_SET(grab_fd, &fds);

   /* Maybe a little nicer to the cpu ? */
   select(grab_fd + 1, &fds, NULL, NULL, NULL); 

   if (ioctl(grab_fd, VIDIOCMCAPTURE, &grab_buf) == -1) {
      perror("ioctl VIDIOCMCAPTURE");
      return;
   }

   if (ioctl(grab_fd, VIDIOCSYNC, &i) == -1) {
      perror("ioctrl VIDIOCSYNC");
      return;
   }
   return;
}

int
init2()
{
	int fd;
	int i;
	grab_data=malloc(_YUV_WIDTH*_YUV_HEIGHT*2);
   i2=malloc(sizeof(struct plop));
   if(i2==NULL)
   {
	   printf("Couldnt allocate memory.\n");
	   exit(1);
   }

//	i2->usb = open("/dev/hc08_00", O_RDWR);		// for 2.4
	i2->usb = open("/dev/hc08_0", O_RDWR);	// for 2.6
	if(i2->usb<0)
	{
		perror("Could not open hc08 device file.");
		exit(-1);
	}
	memset(i2->buf,0x00,SIZE);
	if(write(i2->usb,i2->buf,SIZE)!=SIZE)
		printf("Could not mrite bytes to hc08\n");
	read(i2->usb,i2->buf_,SIZE);
	usleep(_USB_SLEEP_TIME_);
	i2->buf[5]=0x80;
	i2->buf[6]=0x70;
	i2->buf[7]=0x68;
	if(write(i2->usb,i2->buf,SIZE)!=SIZE)
		printf("Could not mrite bytes to hc08\n");
	read(i2->usb,i2->buf_,SIZE);
	usleep(_USB_SLEEP_TIME_);
	i2->lock=0;
	//	i2->tab1=(int **)malloc(yuv_width*yuv_height*sizeof(int));
	//	i2->tab2=(int **)malloc(yuv_width*yuv_height*sizeof(int));
	i2->e_filter=1;
	i2->filter_gb=0;
	i2->filter_gr=0;
	i2->view=0;
	i2->noise_f=0;
	i2->s_t=20;
	i2->sleep=0;
	i2->search=0;
	i2->edge_factor=2;
	i2->flag=0;
	i2->show=0;
	for(i=0;i<PRECOS_SIZE;i++)
	{
		coss[i]=cosf((float)i);
		sinn[i]=sinf((float)i);
	}
	// creating black noise picture
	fd=open(_BLACK_FILE,O_RDONLY);
	if(fd<0)
	{
fprintf(stderr,"Could not open black noise file ... did you create it ?\
to create it just add the -b option as first argument.\n");
		exit(1);
	}
	if(read(fd,i2->noise,320*480)!=320*480)
	{
		fprintf(stderr,"Black noise file corrupt - recreate it !\n");
		exit(1);
	}
	film=0;
	printf("\n\n\
All refered keys are fr_FR keyboard;\n\
Press on Space to activate the mouse, and on 1-2 to change the filter\
(edge detector width).\n\
press A to get original pic, Z for strength of edges, E for phase\n\
Q for switching Green/Blue filter,\n\
S for switching Green/Red filter.\n\
W for switching black noise computation\n\
X for showing the current search statistics\n\
M far switching the 'flag'\n\
H far switching the Hough flag\n\
K for taking one shot\n\
J allow disallow to take a shot every frame\n\
press arrows to change the edge detector factor\n\n\
Pass argument -b to create black noise image.\n\
Please wait 5s before the stream start ...\n\n");
	return(0);
}

	int
conv1(int x, int y)	// make convolution with 3 pixels filters
{
	int i, j;
	int sum_h;
	int sum_v;
	sum_h=0;
	sum_v=0;
	for(i=0;i<=_F_;i++)
		for(j=0;j<=_F_;j++)
		{
			sum_h+=filter1_h[i][j]*i2->tab1[x+i-_H_][y+j-_H_];
			sum_v+=filter1_v[i][j]*i2->tab1[x+i-_H_][y+j-_H_];
		}
	i2->G_h[x][y]=sum_h;
	i2->G_v[x][y]=sum_v;
	if(i2->filter_gb)
	{
		sum_h=0;
		sum_v=0;
		for(i=0;i<=_F_;i++)
			for(j=0;j<=_F_;j++)
			{
		sum_h+=filter1_h[i][j]*i2->tab1_gb[(x+i)/2-_H_][(y+j)/2-_H_];
		sum_v+=filter1_v[i][j]*i2->tab1_gb[(x+i)/2-_H_][(y+j)/2-_H_];
			}
		i=i2->G_h[x][y];
		j=i2->G_v[x][y];
		i2->G_h[x][y]+=abs(sum_h)*(i?i/(abs(i)):1);
		i2->G_v[x][y]+=abs(sum_v)*(j?j/(abs(j)):1);
	}
	if(i2->filter_gr)
	{
		sum_h=0;
		sum_v=0;
		for(i=0;i<=_F_;i++)
			for(j=0;j<=_F_;j++)
			{
		sum_h+=filter1_h[i][j]*i2->tab1_gr[(x+i)/2-_H_][(y+j)/2-_H_];
		sum_v+=filter1_v[i][j]*i2->tab1_gr[(x+i)/2-_H_][(y+j)/2-_H_];
			}
		i=i2->G_h[x][y];
		j=i2->G_v[x][y];
		i2->G_h[x][y]+=abs(sum_h)*(i?i/(abs(i)):1);
		i2->G_v[x][y]+=abs(sum_v)*(j?j/(abs(j)):1);
	}
	return(0);
}

	int
conv2(int x, int y)
{	// make convolution with 5 pixels filters
	int i, j;
	int sum_h;
	int sum_v;
	sum_h=0;
	sum_v=0;
	for(i=0;i<=_F_;i++)
		for(j=0;j<=_F_;j++)
		{
			sum_h+=filter2_h[i][j]*i2->tab1[x+i-_H_][y+j-_H_];
			sum_v+=filter2_v[i][j]*i2->tab1[x+i-_H_][y+j-_H_];
		}
	i2->G_h[x][y]=sum_h;
	i2->G_v[x][y]=sum_v;
	if(i2->filter_gb)
	{
		sum_h=0;
		sum_v=0;
		for(i=0;i<=_F_;i++)
			for(j=0;j<=_F_;j++)
			{
		sum_h+=filter2_h[i][j]*i2->tab1_gb[x/2+i-_H_][y/2+j-_H_];
		sum_v+=filter2_v[i][j]*i2->tab1_gb[x/2+i-_H_][y/2+j-_H_];
			}
		i=i2->G_h[x][y];
		j=i2->G_v[x][y];
		i2->G_h[x][y]+=abs(sum_h)*(i?i/(abs(i)):1);
		i2->G_v[x][y]+=abs(sum_v)*(j?j/(abs(j)):1);
	}
	if(i2->filter_gr)
	{
		sum_h=0;
		sum_v=0;
		for(i=0;i<=_F_;i++)
			for(j=0;j<=_F_;j++)
			{
		sum_h+=filter2_h[i][j]*i2->tab1_gr[x/2+i-_H_][y/2+j-_H_];
		sum_v+=filter2_v[i][j]*i2->tab1_gr[x/2+i-_H_][y/2+j-_H_];
			}
		i=i2->G_h[x][y];
		j=i2->G_v[x][y];
		i2->G_h[x][y]+=abs(sum_h)*(i?i/(abs(i)):1);
		i2->G_v[x][y]+=abs(sum_v)*(j?j/(abs(j)):1);
	}
	return(0);
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

//                      Random GHT

int
random_(int *x, int *y)
{
   int i,j, z;
   z=0;
   while(1)
   {
   i=rand()%_YUV_WIDTH;
   j=rand()%_YUV_HEIGHT;
   if(i2->S[i][j])
   break;
   if(z>_MAX_EDGE_ERROR_)
   return(1);
   }
 *x=i;
 *y=j;
 return(0);
 }

int
RGHT()
{
	int i;
	int x1, y1, x2, y2;
	float sx, sy;
	float ox, oy;
printf("PLEASE UPDATE RGHT() BEFORE USING IT\n\
READ THE HEADER OF EGHT PART\n");
exit(1);
	for(x1=0;x1<yuv_width;x1++)
	for(y1=0;y1<yuv_height;y1++)
		i2->result[x1][y1]=0;
	for(i=0;i<_YUV_WIDTH*_YUV_HEIGHT*1;i++)
	{
		if(random_(&x1,&y1) || random_(&x2,&y2))
		return(1);
// good luck to the one who tries to understand what I wrote :
sx=(float)(x2-x1)/(float)(i2->R_x[(int)(i2->P[x1][y1]*_R_RES_/PI)]-
i2->R_x[(int)(i2->P[x2][y1]*_R_RES_/(2*PI))]);
sy=(float)(y2-y1)/(float)(i2->R_y[(int)(i2->P[x1][y1]*_R_RES_/PI)]-
i2->R_y[(int)(i2->P[x2][y1]*_R_RES_/(2*PI))]);
//ox=((float)x1+sx*i2->R_x[(int)(i2->P[x1][y1]*_R_RES_/(2*PI))] + // to update
//(float)x2+sx*i2->R_x[(int)(i2->P[x2][y1]*_R_RES_/(2*PI))])/2;
//oy=((float)y1+sy*i2->R_y[(int)(i2->P[x1][y1]*_R_RES_/(2*PI))] + // to update
//(float)y2+sy*i2->R_y[(int)(i2->P[x2][y1]*_R_RES_/(2*PI))])/2;
// if you really think you get it, just email me to let me know ...
	if(	// is O inside the picture ?
		((ox>=0)&&(ox<_YUV_WIDTH)&&(oy>=0)&&(oy<_YUV_HEIGHT))
		&&	// is the scaling ratio acceptable ?
		((sx>0.5)&&(sx<2)&&(sy>0.5)&&(sy<2))
	)
		i2->result[(int)ox][(int)oy]=0xFF;
	}
	return(0);
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

//                     Exhaustive GHT

// when this part was first written, i2.R_table only had one point per radius
// ( [_R_WIDTH_] did not exist yet )
// if you now want to use this, you have to modify compute_local_r() either
// to use i2.R_Table with all the width, or make the local R-table use only the 
// first points of i2.R-table

int
moy_local(int *x, int *y) // put the cursor at the average center of the object
{
   int i;
   float m, n;
   m=0; n=0;
   for(i=0;i<_R_RES_;i++)
   {
//	m+=i2->R_x[i];	// to update
//	n+=i2->R_y[i];	// to update
   }
   m/=_R_RES_;
   n/=_R_RES_;

   if(tmp_i)
   printf("	m=%f_n=%f\n",m,n);
 *x += m;
 *y += n;
 return(0);
 }

int
compute_local_r(int _x, int _y)
{
	int r;
	int i, j;
	int count;		// number of valid fields in the R-Table
	float x, y;
	int o;
	int exit;
	int plop;		// radius when half ot points have been found
 
printf("PLEASE UPDATE compute_local_r() BEFORE USING IT\n\
READ THE HEADER OF EGHT PART\n");
return(1);
//exit(0);
	for(i=0;i<_R_RES_;i++)
	{
		plaf.R_valid[i]=0;
		plaf.R_s[i]=0;
		plaf.R[i]=0;
	}
	r=0;
	count=0;
	exit=0;
	plop=_YUV_WIDTH;
	while(!exit)
	{
		r++;	// increase circle
		for(o=0;o<_R_RES_;o++)	// walk along the whole circle
		{	// compute absolute coordinotes or current point
			x=_x+(float)r*cos((float)o);
			y=_y+(float)r*sin((float)o);
			//	x=_x+(float)r*coss[o];
			//	y=_y+(float)r*sinn[o];
			if(x<0 || y<0 || x>_YUV_WIDTH || y>240)
				continue;
if( ((i2->W[(int)x][(int)y]>(i2->edge_factor+2))&&(plaf.R_valid[o]<=(r/4)))||
	( (i2->W[(int)x][(int)y]>i2->edge_factor) && (!plaf.R_valid[o]) ) 
)
		// compare curent edge with the one just nearer the center
			{
				plaf.R[o]=r;
				if(!plaf.R_valid[o])
				count++;
				plaf.R_valid[o]+=5;
				plaf.R_s[o]=i2->S[(int)x][(int)y];
				plaf.R_x[o]=(int)x-_x;
				plaf.R_y[o]=(int)y-_y;
			}
if( (i2->W[(int)x][(int)y]>i2->edge_factor) && (plaf.R_valid[o]) )
				plaf.R_valid[o]++;
		}
		if((count>=(_R_RES_*0.7))&&(!plop))
		plop=r;
		if((count>(_R_RES_*0.9))||(r>130)||(r>plop*3))
		exit=1;
	}
	return(0);
}

int
compare_local_to_R()
{
	int i, p;
	float r, s;
	r=0; p=0;
	for(i=0;i<_R_RES_;i++)
	{
		if(plaf.R_valid[i] && i2->R_valid[i])
		{	// check if point is member of the edge
		//	s=100*(1-(float)plaf.R[i]/(float)i2->R[i]); // to update
			if(s<0)
				s=-s;
			if(s<_ACCEPTED_RATE_)
			{
				r+=s;
				p++;
			}
		}
	}
	if(100*(float)p/(float)_R_RES_<_ACCEPTED_RATE__)
		return(999);
	r/=(float)p;
	return(r);
}

/////////////////////// TO USE EGHT, PUT _R_RES_ = 10

	int
EGHT() // exhaustive GHT ( we scan the picture )
{
	// I run over the whole picture and consider succevly each point as
	// an anchor point, and then compute the local R-Table ... if it 
	// matches "THE" i2->R-Table then we get the good point.
	int i, j;
	int x,y;
	int tmp;

	for(i=0;i<yuv_width;i++)
		for(j=0;j<yuv_height;j++)
			i2->result[i][j]=0;

	for(i=_H2_;i<yuv_width-_H2_;i++)
		for(j=_H2_;j<yuv_height-_H2_;j++)
		{
			if(i2->S[i][j])
				i2->W[i][j]=length(i,j);
			else
				i2->W[i][j]=0;
		}

	for(i=0;i<_YUV_WIDTH;i+=2)
	{
		if(!(i%2))
			printf("i=%3.1f\%\n",100*(float)i/(float)_YUV_WIDTH);
		for(j=0;j<_YUV_HEIGHT;j+=2)
		{
			if((abs(x-tmp_x)<4)&&(abs(y-tmp_y)<4))
				tmp_i=1;
			else
				tmp_i=0;
			compute_local_r(i,j);
			// compute R_table of cur object
			x=i; y=j;
			//	moy_local(&x,&y);	// find average center
			//	compute_local_r(x,y,R,R_valid,R_x,R_y);
			// recompute R_table
			tmp=compare_local_to_R();
			if(tmp<_ACCEPTED_RATE_)
				i2->result[x][y]=0xFF;
		}
	}
	i2->show=0;
	return(0);
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

//                          Phase and edge based GHT

int
PGHT(int *xx, int *yy)		// phase shape GHT
{
	int i, j;
	int x,y;
	int tmp;
	int p;
	float o_p, o_w;

	int	mat[7][7]={
		{-0, -0, -0, -0, -0, -0, -0},
		{-0, -1, -1, -1, -1, -1, -0},
		{-0, -1,  2,  2,  2, -1, -0},
		{-0, -1,  2,  4,  2, -1, -0},
		{-0, -1,  2,  2,  2, -1, -0},
		{-0, -1, -1, -1, -1, -1, -0},
		{-0, -0, -0, -0, -0, -0, -0},
	};
	int result[_YUV_WIDTH][_YUV_HEIGHT];
	int max_r, max_x, max_y;
	// location where the filter will detect a maxima

	for(i=0;i<yuv_width;i++)
		for(j=0;j<yuv_height;j++)
			i2->result[i][j]=0;

	for(i=_H2_;i<yuv_width-_H2_;i++)
		for(j=_H2_;j<yuv_height-_H2_;j++)
		{
			if(i2->S[i][j])
				i2->W[i][j]=length(i,j);
			else
				i2->W[i][j]=0;
		}

	for(i=0;i<_YUV_WIDTH;i+=1)
		for(j=0;j<_YUV_HEIGHT;j+=1)
		{
			int w;
			if(!i2->S[i][j])
				continue;
			o_p=i2->P[i][j];
			o_w=i2->W[i][j];
			for(p=0;p<_R_RES_;p++)
			for(w=0;w<_R_WIDTH_;w++)
			{
				int x = i - i2->R_x[p][w];
				int y = j - i2->R_y[p][w];
				if(!i2->R_valid[p])
					continue;
		// test if phase in pic is close to any phase in R-Table
//				if(abs(o_p-i2->R_P[p][w])<5)
//				if(abs(o_w-i2->R_W[p][w])<5)
//if((i-i2->R_x[p][w]>=0)&&(i-i2->R_x[p][w]<_YUV_WIDTH)&&
//(j-i2->R_y[p][w]>=0)&&(j-i2->R_y[p][w]<_YUV_HEIGHT))
//	i2->result[i-i2->R_x[p][w]][j-i2->R_y[p][w]]+=0x1;

				if (
					x >= 0 && x < _YUV_WIDTH
					&& y >= 0 && y < _YUV_HEIGHT
					&& abs(o_p - i2->R_P[p][w]) < 5
					&& abs(o_w - i2->R_W[p][w]) < 5
				   )
					++i2->result[x][y];
			}
		}
	max_r=0;
	// point detector filter
	for(i=3;i<_YUV_WIDTH-3;i+=1)	// along the picture
		for(j=3;j<_YUV_HEIGHT-3;j+=1)
		{
			tmp=0;	
			for(x=0;x<7;x++)	// along the matrix
			for(y=0;y<7;y++)	// ( the Matrix has you )
			{
				tmp+=i2->result[i+x-3][j+y-3]*mat[x][y];
			}
			result[i][j]=tmp;
			if(tmp>max_r)
			{
				max_r=tmp;
				max_x=i;
				max_y=j;
			}
		}

	*xx=max_x;
	*yy=max_y;

	return(0);
}

//				End of GHT algos

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

int
run2(long frames)
{
	int x, y, w;
	int i,j;
	float r;

	if(i2->noise_f)
		for(x=0;x<_YUV_WIDTH*_YUV_HEIGHT+320*120*2;x++)
		{
			if(grab_data[x]>i2->noise[x] || i2->flag)
grab_data[x]-=i2->noise[x]+(x>=_YUV_WIDTH*_YUV_HEIGHT?0x80:0);
			else
grab_data[x]=i2->noise[x]-grab_data[x]+(x>=_YUV_WIDTH*_YUV_HEIGHT?0x80:0);
		}
	for(x=0;x<yuv_width;x++)	// generate light array
		for(y=0;y<yuv_height;y++)
			i2->tab1[x][y]=(int)grab_data[x+y*yuv_width];
	for(x=0;x<yuv_width/2;x++)	// color orrays
		for(y=0;y<yuv_height/2;y++)
		{
i2->tab1_gb[x][y]=(int)grab_data[x+y*yuv_width/2+_YUV_WIDTH*_YUV_HEIGHT];
i2->tab1_gr[x][y]=(int)grab_data[x+y*yuv_width/2+320*240+160*120];
		}
	switch(i2->e_filter)	// apply edge filter
	{
		case 1:
			for(x=_F_;x<yuv_width-_F_;x++)
				for(y=_F_;y<yuv_height-_F_;y++)
					conv1(x,y);
			break;
		case 2:
			for(x=_F_;x<yuv_width-_F_;x++)
				for(y=_F_;y<yuv_height-_F_;y++)
					conv2(x,y);
			break;
	}
	for(x=0;x<yuv_width;x++)
		for(y=0;y<yuv_height;y++)
		{	//compute strength and phase
			i2->S[x][y]=0;
			r=sqrt((float)(i2->G_h[x][y])*(float)(i2->G_h[x][y])+
				(float)(i2->G_v[x][y])*(float)(i2->G_v[x][y]));
			if(r>i2->s_t)
			{
				i2->S[x][y]=r;
r=atanf((float)(i2->G_v[x][y])/(float)(i2->G_h[x][y]))+1*PI/2;
			}
			if(r<0)
				r+=PI;
			//	if(r>=PI)
			//		r-=PI;
			i2->P[x][y]=r;
		}
		switch(i2->view)
		{
		case 0:	// show real image
			break;
		case 1:	// shom strength
	for(x=0;x<yuv_width;x++)	// this is the shown picture
		for(y=0;y<yuv_height;y++)
		{
			r=i2->S[x][y];
			if(r<0x10)
				r=0;
			if(r>0xFF)
				r=0xFF;
			grab_data[x+y*yuv_width]=(unsigned char)r;
		}
			break;
		case 2:	// show phase
	for(x=0;x<yuv_width;x++)	// this is the shown picture
		for(y=0;y<yuv_height;y++)
		{
			r=1;
			if(i2->S[x][y])
				r=i2->P[x][y]*255/(2*PI);
			if(r<0)
				r=0;
			if(r>0xFF)
				r=0xFF;
			grab_data[x+y*yuv_width]=(unsigned char)r;
		}
			break;
		case 3:	// show W
	for(x=0;x<yuv_width;x++)	// this is the shown picture
		for(y=0;y<yuv_height;y++)
		{
			r=i2->W[x][y];
			r=r*255/(_H2_*2+1);
		grab_data[x+y*yuv_width]=(unsigned char)r;
		}
			break;
		case 4:	// show the Hough space
	{	// show the Hough table
//		int x, y;
		int res=100;
		int tab[(int)(res*PI*2)];
		int c[(int)(res*PI*2)];
		int max, val;
//		int max, cur;
//		unsigned char cur, max;
		val=0; max=0;
		for(x=0;x<_YUV_WIDTH;x++)
		{
			tab[x]=0;
			c[x]=0;
			for(y=0;y<_YUV_HEIGHT;y++)
				grab_data[x+y*yuv_width]=0;
		}
		for(x=0;x<_YUV_WIDTH;x++)
		for(y=0;y<_YUV_HEIGHT;y++)
		{
			float k;
			float a_, a, r;
			int tmp1, tmp2;
			if(i2->S[x][y]==0)
				continue;
/*			a_=atanf((float)y/(float)x);
			for(k=0;k<320;k++)
			{
				a=k*PI/320+a_;
				a=(a+PI/2)*320/(32*PI);
				if(a<0)
					printf("%f\n",a);
				if(a>320)
					printf("%f\n",a);
				r=sqrtf((float)(x*x+y*y))*cosf(i2->P[x][y])/2;
				if(grab_data[(int)(a+r*yuv_width)]<254)
					grab_data[(int)(a+r*yuv_width)]++;
			}*/
	//		printf("p=%f s=%f\n", i2->P[x][y],i2->S[x][y]);
			grab_data[
				(int)
				(i2->P[x][y]*res)
				+
				(int)
				(i2->S[x][y]*2)*yuv_width
				]+=10;
			tmp1=(int)(i2->P[x][y]*res);
			tmp2=(int)(i2->S[x][y]*sqrtf(i2->S[x][y]));
			if(tmp1>(int)(res*PI*2))
				continue;
			tab[tmp1]+=tmp2;
			c[tmp1]++;
	/*		if((val>3000)||(tmp2>100)||(tmp1>400))
			{
	printf("x=%i y=%i tmp1=%i tmp2=%i - P=%f S=%f - max=%i val=%i\n",
						x,y,tmp1,tmp2,
						i2->P[x][y], i2->S[x][y],
						max, val);
				for(x=0;x<320;x++)
					printf("%3.3i:%5.5i/%3.3i ",x,tab[x],c[x]);
				exit(1);
			}*/
			if(tab[tmp1]>val)
			{
				max=tmp1;
				val=tab[max];
			}
		//	grab_data[(int)(
		//		(i2->P[x][y]*_YUV_WIDTH/PI)
		//		+
		//		(i2->S[x][y]*_YUV_HEIGHT/56)*yuv_width
		//		)]+=10;
		}
	//	printf("max=%i val=%i\n",max, val);
		grab_data[max]=0xFF;
	grab_data[max/2+(0*yuv_width/2)+320*240+160*120]=0xFF;
	//	for(x=0;x<_YUV_WIDTH;x++)
	//	for(y=0;y<_YUV_HEIGHT;y++)
	}
			break;
		case 5:
//	if(!i2->sleep)	// show the accumulated results
		// exept if just selected the object (???)
	{	for(x=0;x<yuv_width;x++)
			for(y=0;y<yuv_height;y++)
			{
	grab_data[x+y*yuv_width]=(unsigned char)i2->result[x][y];
	//			if(i2->result[x][y])
	//			{
	//			grab_data[x+y*yuv_width]=0xFF;
	//			grab_data[x/2+y*yuv_width/4+320*240]=0xFF;
	//			}
	//			else
	//			grab_data[x+y*yuv_width]=0x00;
			}
	}
			break;
		default:
			printf("The switch selector of the show part went wrong\n");
			break;
		}
	if(i2->sleep)	// draw in red the selected object only once
	{
		for(x=0;x<_R_RES_;x++)
		for(w=0;w<_R_WIDTH_;w++)
		{
			if(i2->R_valid[x])
				grab_data[
					(i2->R_x[x][w]+i2->mouse_x)/2+
					((i2->R_y[x][w]+i2->mouse_y)/2)*yuv_width/2+
					320*240+160*120
					]=0xFF;
		}
		scrot();
	}
	else if(i2->search) // otherwise search the pattern in the picture.
	{
		struct timeval t_start, t_end;
// call one GHT
		gettimeofday(&t_start,NULL);
		//		RGHT();
		//		EGHT();
		PGHT(&x,&y);
		gettimeofday(&t_end,NULL);
		printf("PGHT: %fms\n",time_diff(&t_start,&t_end));
		printf("x=%3.3i y=%3.3i\n",x,y);
		for(i=-1;i<=1;i++)
			for(j=-1;j<=1;j++)
			{
				if(i || j )
	grab_data[x/2+i+((y/2+j)*yuv_width/2)+320*240+160*120]=0xFF;
				else
	grab_data[x/2+i+((y/2+j)*yuv_width/2)+320*240+160*120]=0x00;
			}
		grab_data[x/2+y*yuv_width/4+320*240]=0xFF;

		i2->buf[5]+=(_YUV_WIDTH/2-x)/40;
		i2->buf[6]-=(_YUV_HEIGHT/2-y)/40;

		if(i2->buf[5]<_MIN_5_)
			i2->buf[5]=_MIN_5_;
		if(i2->buf[5]>_MAX_5_)
			i2->buf[5]=_MAX_5_;
		if(i2->buf[6]<_MIN_6_)
			i2->buf[6]=_MIN_6_;
		if(i2->buf[6]>_MAX_6_)
			i2->buf[6]=_MAX_6_;
		if(write(i2->usb,i2->buf,SIZE)!=SIZE)
			printf("Could not mrite bytes to hc08\n");
		read(i2->usb,i2->buf_,SIZE);
		usleep(_USB_SLEEP_TIME_);
	}
	return(0);
}

	int
length(int x, int y)
{
	// here I determine the shape of the parallel line crossing the
	// current point, and count how many points have a close edge strength
	int r;
	int i;
	r=0;
	for(i=-_H2_;i<_H2_;i++) // walk along 10 pixels
		if(abs(
			i2->S[ x+(int)(i*(sinf(i2->P[x][y]-(2*PI)/2))) ]
			[ y+(int)(i*(cosf(i2->P[x][y]-(2*PI)/2))) ] -
			((int)i2->S[x][y])
		      )
		<10)
			// quality factor : lower makes filter more restrictive
			r++;
	return(r);
}

	int
moy_r()	// put the cursor at the average center of the object
{
	int i, w;
	float m, n;
	m=0; n=0;
	for(i=0;i<_R_RES_;i++)
	for(w=0;w<_R_WIDTH_;w++)
	{
		m+=i2->R_x[i][w];
		n+=i2->R_y[i][w];
	}
	m/=_R_RES_;
	n/=_R_RES_;
	i2->mouse_x+=m;
	i2->mouse_y+=n;
	return(0);
}

int
extract_R(int mouse_x, int mouse_y)
{
	int r, w;
	int R_s[_R_RES_][_R_WIDTH_];	// strength of edge for the R-Table
	int i, j;
	int count;		// number of valid fields in the R-Table
	float x, y;
	int o;
	int exit;
	int plop;		// radius when half ot points have been found
	i2->mouse_x=mouse_x;
	i2->mouse_y=mouse_y;
	tmp_x=mouse_x; tmp_y=mouse_y;
	for(i=0;i<_R_RES_;i++)
	for(w=0;w<_R_WIDTH_;w++)
	{
		i2->R_valid[i][w]=0;
		R_s[i][w]=0;
		i2->R[i][w]=0;
	}
	for(i=_H2_;i<yuv_width-_H2_;i++)
		for(j=_H2_;j<yuv_height-_H2_;j++)
		{
			if(i2->S[i][j])
				i2->W[i][j]=length(i,j);
			else
				i2->W[i][j]=0;
		}
	r=0;
	count=0;
	exit=0;
	plop=_YUV_WIDTH;
	while(!exit)
	{
		r++;	// increase circle
		for(o=0;o<_R_RES_;o++)	// walk along the whole circle
		{	// compute absolute coordinates or current point
			x=mouse_x+(float)r*cosf((float)o);
			y=mouse_y+(float)r*sinf((float)o);
			if(x<0 || y<0 || x>_YUV_WIDTH || y>240)
				continue;
			// determine location of first available cell
			for(w=0;(w<_R_WIDTH_)&&(i2->R_valid[o][w]!=0);)
				w++;
			if(w>=_R_WIDTH_)
				continue;
if( ((i2->W[(int)x][(int)y]>(i2->edge_factor+2))&&(i2->R_valid[o][w]<=(r/4)))||
	( (i2->W[(int)x][(int)y]>i2->edge_factor) && (!i2->R_valid[o][w]) ) 
  )
//			if(((i2->S[(int)x][(int)y]>R_s[o])&&(!i2->R_valid[o]))||
//		(i2->S[(int)x][(int)y]>((R_s[(int)o])+2*(r-i2->R[(int)o]))))
// compare curent edge with the one just nearer the center
			{
				i2->R[o][w]=r;
				if(!i2->R_valid[o][w])
					count++;
				i2->R_valid[o][w]+=5;
				R_s[o][w]=i2->S[(int)x][(int)y];
				i2->R_x[o][w]=(int)x-mouse_x;
				i2->R_y[o][w]=(int)y-mouse_y;
			}
	if( (i2->W[(int)x][(int)y]>i2->edge_factor) && (i2->R_valid[o][w]) )
				i2->R_valid[o][w]++;
			i2->R_P[o][w]=i2->P[(int)x][(int)y];
			i2->R_W[o][w]=i2->W[(int)x][(int)y];
		}
		if((count>=(_R_RES_*_R_WIDTH_*0.7))&&(!plop))
			plop=r;
		if((count>(_R_RES_*_R_WIDTH_*0.9))||(r>130)||(r>plop*3))
			exit=1;
	}
	i2->sleep=1;
	i2->search=1;
	i2->show=0;
	return(0);
}

	int
save_black_noise()
{
	int fd;
	long int i,j;
	int tab[320*480];
	unsigned char av[320*480];
	fd=open(_BLACK_FILE,O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
	if(fd<0)
	{
		fprintf(stderr,"Could not open %s.\n",_BLACK_FILE);
		return(1);
	}
	printf("Recording black noise picture ... takes about 8s\n");
	for(i=0;i<30;i++)	// wait until the camera self calibrated
	{
		grab_image();
	}
	for(j=0;j<320*480;j++)
		tab[j]=0;
	for(i=0;i<10;i++)	// wait until the camera self calibrated
	{
		grab_image();
		for(j=0;j<320*480;j++)
			tab[j]+=(long int)grab_data_tmp[j];
	}
	for(j=0;j<320*480;j++)
		av[j]=(unsigned char)(tab[j]/10);
	write(fd,av,320*480);
printf("Black noise picture taken. Dont take care about the next segfault.\n");
	return(0);
}

int
child()
{
	if(i2->search)
		usleep(45*1000);
//		gettimeofday(&t_start,NULL);
	grab_image();
//		gettimeofday(&t_end,NULL);
//		printf("GRAB: %fms\n",time_diff(&t_start,&t_end));
	return(0);
}

int
scrot()
{
	unsigned char (*data)[3];
	int w, h;
	char fname[1024];
	FILE *fp;
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	png_bytep *row_pointers = NULL;
	int j;

	time_t t;
	struct tm *tm;
	
	w=_YUV_WIDTH;	// Width of the picture - from constantes
	h=_YUV_HEIGHT;	// Height of the pic
	time(&t);
	tm=localtime(&t);
	strcpy(fname,"xvcam-shot-YYYY_MM_DD-HH_MM_SS.png\0");
	sprintf(&fname[11],"%4.4i_%2.2i_%2.2i-%2.2i_%2.2i_%2.2i.png\0",
		tm->tm_year+1900,tm->tm_mon+1,tm->tm_mday,
		tm->tm_hour,tm->tm_min,tm->tm_sec);
	
	printf("The shot will be stored in %s\n",fname);
	fp=fopen(fname, "wb");
	if (fp == NULL)
	{
		printf("Could not create shot file in current directory");
		return(1);
	}

	(data) = malloc(w * h * sizeof *data);
	if (data == NULL)
	{
		printf("Could not allocate memory for PNG export.\n");
		return(1);
	}
	{
		int i, j;
		unsigned char c;
		for(i=0;i<w;i++)
			for(j=0;j<h;j++)
			{
				float y, u, v, r, g, b;
// grab_data[] is the place where is stored YUV picture
				y=grab_data[i+j*w];
				u=grab_data[i/2+(j/2)*w/2+w*h]-128;
				//u=0;
				v=grab_data[i/2+(j/2)*w/2+w*h+w*h/4]-128;
				//v=0;
				r=y+(v)*1.4023;
				g=y-(u)*0.3437-(v)*0.748;
				b=y+(u)*1.7734;
				if(r<0) r=0;
				if(g<0) g=0;
				if(b<0) b=0;
				if(r>255) r=255;
				if(g>255) g=255;
				if(b>255) b=255;
				(*((data)+1*(j*w+i)))[0]=r;
				(*((data)+1*(j*w+i)))[1]=g;
				(*((data)+1*(j*w+i)))[2]=b;
			}
	}

	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	if (png_ptr == NULL)
		goto err_png;
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL)
		goto err_png;
	if (setjmp(png_jmpbuf(png_ptr)))
		goto err_png;
	
	png_init_io(png_ptr, fp);
	
	png_set_IHDR(png_ptr, info_ptr, w, h,
			8, PNG_COLOR_TYPE_RGB,
			PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
			PNG_FILTER_TYPE_DEFAULT);
//	png_set_compression_level(png_ptr,10);
	png_write_info(png_ptr, info_ptr);
	
	row_pointers = png_malloc(png_ptr, h * sizeof *row_pointers);
	for (j=0; j < h; ++j)
		row_pointers[j] = (png_byte*)data[j*w+0];

	png_write_image(png_ptr, row_pointers);
	png_free(png_ptr, row_pointers);
	png_write_end(png_ptr, NULL);
	png_destroy_write_struct(&png_ptr, &info_ptr);
	fclose(fp);
	return 0;

err_png:
	if (row_pointers) png_free(png_ptr, row_pointers);
	png_destroy_write_struct(&png_ptr, &info_ptr);
	fclose(fp);
	return 2;
	
}

int
main(int argc, char **argv)
{
	int xv_port = -1, i, d, screen, CompletionType;
	unsigned int ud, width, height;
	long frames;
	unsigned int ver, rel, req, ev, err, adapt;
	long stack[1024];
	Display *dpy;
	Window window, _dw;
	XSizeHints hint;
	XSetWindowAttributes xswa;
	XWindowAttributes attribs;
	XVisualInfo vinfo;
	XEvent event;
	GC gc;
	XvAdaptorInfo *ai;
	XvImage *yuv_image;
	XShmSegmentInfo yuv_shminfo;
	Atom wmDeleteWindow;

	dpy = XOpenDisplay(NULL);
	screen = DefaultScreen(dpy);

	XGetWindowAttributes(dpy, DefaultRootWindow(dpy), &attribs);
	XMatchVisualInfo(dpy, screen, attribs.depth, TrueColor, &vinfo);
	wmDeleteWindow = XInternAtom(dpy, "WM_DELETE_WINDOW", False);

	if(argc!=1)
		if(!strcmp(argv[1],"-b"))
		{
			init();
			save_black_noise();
			return(0);
		}

	frames=0;

	hint.x = 1;
	hint.y = 1;
	hint.width = yuv_width;
	hint.height = yuv_height;
	hint.flags = PPosition | PSize;

	xswa.colormap = XCreateColormap(dpy, DefaultRootWindow(dpy), vinfo.visual, 
			AllocNone);
	xswa.event_mask = StructureNotifyMask | ExposureMask;
	xswa.background_pixel = 0;
	xswa.border_pixel = 0;

	window = XCreateWindow(dpy, DefaultRootWindow(dpy), 0, 0,
			yuv_width*ASPECT_RATIO, yuv_height*ASPECT_RATIO, 0,
			vinfo.depth, InputOutput, vinfo.visual, 
			CWBackPixel | CWBorderPixel | CWColormap | 
			CWEventMask, &xswa);

	//   XSelectInput(dpy, window, StructureNotifyMask);
	XSelectInput(dpy, window, StructureNotifyMask
		| ExposureMask | ButtonPressMask | PointerMotionMask |
		ButtonMotionMask | KeyPressMask | KeyReleaseMask);
	XSetStandardProperties(dpy, window, "xvcam", "xvcam", None, NULL, 0,
			&hint);

	XSetWMProtocols(dpy, window, &wmDeleteWindow, 1);
	XMapWindow(dpy, window);

	if (XShmQueryExtension(dpy))
		CompletionType = XShmGetEventBase(dpy) + ShmCompletion;
	else
		exit(-1);

	if (Success !=
			XvQueryExtension(dpy, &ver, &rel, &req, &ev, &err))
		fprintf(stderr, "Couldn't do Xv stuff\n");

	if (Success !=
			XvQueryAdaptors(dpy, DefaultRootWindow(dpy), &adapt, &ai))
		fprintf(stderr, "Couldn't do Xv stuff\n");

	for (i = 0; i < (int) adapt; i++) {
		xv_port = ai[i].base_id;
	}

	if (adapt > 0)
		XvFreeAdaptorInfo(ai);

	gc = XCreateGC(dpy, window, 0, 0);

	yuv_image = XvShmCreateImage(dpy, xv_port, GUID_I420_PLANAR, 0, yuv_width,
			yuv_height, &yuv_shminfo);
	yuv_shminfo.shmid = shmget(IPC_PRIVATE, yuv_image->data_size, 
			IPC_CREAT | 0777);
	yuv_shminfo.shmaddr = (char *) shmat(yuv_shminfo.shmid, 0, 0);
	yuv_image->data = yuv_shminfo.shmaddr;
	yuv_shminfo.readOnly = False;

	if (!XShmAttach(dpy, &yuv_shminfo)) {
		printf("XShmAttach go boom boom!\n");
		exit(-1);
	}

	init();
	init2();

	memset(yuv_image->data,0x00,320*240);
	memset(320*240+yuv_image->data,0x80,320*240);
	for(frames=0;frames<_SKIP_;frames++)
	{
		int a, b;
		for(a=0;a<255*frames/_SKIP_;a++)
			for(b=110;b<130;b++)
				yuv_image->data[a+b*320]=0x80;
		XGetGeometry(dpy, window, &_dw, &d, &d, &width, &height,
				&ud, &ud);
		XvShmPutImage(dpy, xv_port, window, gc, yuv_image, 0, 0,
				yuv_image->width, yuv_image->height, 0, 0,
				width, height, True);
		grab_image();
	}
	frames=0;

	while (1) 
	{
		struct timeval t_cur, t_old;
		struct timeval t_start, t_end;
		int res;

		gettimeofday(&t_cur,NULL);
		if(frames)
//		printf("Last loop took %f ms.\n",time_diff(&t_old,&t_cur));
		t_old.tv_sec=t_cur.tv_sec;
		t_old.tv_usec=t_cur.tv_usec;
//		gettimeofday(&t_start,NULL);
//		grab_image();
//		gettimeofday(&t_end,NULL);
//		printf("GRAB: %fms\n",time_diff(&t_start,&t_end));

	if(res=clone(child,&stack[1020],SIGCHLD|CLONE_FILES|CLONE_VM,NULL)==-1)
		{
			printf("Could not fork.\n");
			exit(1);
		}
		
		memcpy(grab_data,grab_data_tmp,_YUV_WIDTH*_YUV_HEIGHT*2);
	
		gettimeofday(&t_start,NULL);
		run2(frames);
		if(film)
			scrot();
		gettimeofday(&t_end,NULL);
	//	printf("RUN2: %fms\n",time_diff(&t_start,&t_end));
		
		gettimeofday(&t_start,NULL);
		waitpid(res,NULL,0);
		gettimeofday(&t_end,NULL);
	//	printf("wait: %fms\n\n",time_diff(&t_start,&t_end));
		
		memcpy(yuv_image->data, grab_data, yuv_image->data_size);

		XGetGeometry(dpy, window, &_dw, &d, &d, &width, &height, &ud, &ud);
		XvShmPutImage(dpy, xv_port, window, gc, yuv_image, 0, 0,
			yuv_image->width, yuv_image->height, 0, 0,
			width, height, True);
		frames++;
		XPending(dpy);
		//      if(i2->sleep)
		//	      sleep(2);
		i2->sleep=0;

		while (XPending(dpy))
		{
			XNextEvent(dpy, &event);
			switch(event.type)
			{
				case ClientMessage:
					{
	if (event.xclient.format == 32 &&
	event.xclient.data.l[0] == (signed) wmDeleteWindow)
	exit(0);
					};	break;
				case ButtonPress:
					{
						int button;
						button=3;
						switch(event.xbutton.button)
						{
							case Button1:
								button=1;
								if(i2->lock)
								{
						i2->buf[7]++;
						if(i2->buf[7]<0x10)
							i2->buf[7]=0x10;
						if(i2->buf[7]>0xF0)
							i2->buf[7]=0xF0;
					if(write(i2->usb,i2->buf,SIZE)!=SIZE)
					printf("Could not mrite bytes to hc08\n");
					read(i2->usb,i2->buf_,SIZE);
						usleep(_USB_SLEEP_TIME_);
						i2->search=0;
						i2->show=0;
					}
					else
					{
//			printf("Extracting object from location (%i,%i) ...\n",
//					event.xbutton.x,event.xbutton.y);
	extract_R(event.xbutton.x/ASPECT_RATIO,event.xbutton.y/ASPECT_RATIO);
	moy_r();
	extract_R(i2->mouse_x,i2->mouse_y);
					}
					break;
				case Button2:
					button=2;
					break;
				default: // should be button 3 ( or 4/5 ?)
					if(i2->lock)
					{
						i2->buf[7]--;
						if(i2->buf[7]<0x10)
							i2->buf[7]=0x10;
						if(i2->buf[7]>0xF0)
							i2->buf[7]=0xF0;
				if(write(i2->usb,i2->buf,SIZE)!=SIZE)
					printf("Could not mrite bytes to hc08\n");
				read(i2->usb,i2->buf_,SIZE);
						usleep(_USB_SLEEP_TIME_);
						i2->search=0;
						i2->show=0;
					}
					break;
			}
/*
printf("Button %i pressed at (%i,%i)\n",button,event.xbutton.x,event.xbutton.y);
*/
		};	break;
		case MotionNotify:
		{
	//	printf("position: (%i,%i)\n",event.xbutton.x,event.xbutton.y);
			if(i2->lock)
			{	// move the servos
				// x to z axis
				if(abs(event.xbutton.x-i2->oldx)<10)
					i2->buf[5]+=(event.xbutton.x-i2->oldx)/1;
				if(i2->buf[5]<_MIN_5_)
					i2->buf[5]=_MIN_5_;
				if(i2->buf[5]>_MAX_5_)
					i2->buf[5]=_MAX_5_;
				i2->oldx=event.xbutton.x;
				// y to x axis
				if(abs(event.xbutton.y-i2->oldy)<10)
					i2->buf[6]-=(event.xbutton.y-i2->oldy)/1;
				if(i2->buf[6]<_MIN_6_)
					i2->buf[6]=_MIN_6_;
				if(i2->buf[6]>_MAX_6_)
					i2->buf[6]=_MAX_6_;
				i2->oldy=event.xbutton.y;
				if(write(i2->usb,i2->buf,SIZE)!=SIZE)
					printf("Could not mrite bytes to hc08\n");
				read(i2->usb,i2->buf_,SIZE);
				usleep(_USB_SLEEP_TIME_);
				i2->search=0;
				i2->show=0;
				
			}
		}; break;
		case KeyPress:
		{
			switch(event.xkey.keycode)
			{
			case 9:	// echap plessed ...
				exit(0);
				break;
			case 65:	// space
				i2->lock=!i2->lock;
				i2->oldx=event.xkey.x;
				i2->oldy=event.xkey.y;
				i2->search=0;
				break;
				// dvorak keys : azerty : qwerty
		//	case 19:	// 0
		//		i2->filter=0;
		//		break;
			case 10:	// 1:&:1
				i2->e_filter=1;
				i2->s_t=20;
				break;
			case 11:	// 2::2
				i2->e_filter=2;
				i2->s_t=50;
				break;
			case 24:	// ':a"q
				i2->view=0;
				break;
			case 25:	// ,:z:w
				i2->view=1;
				break;
			case 26:	// .:e:e
				i2->view=2;
				break;
			case 27:	// p:r:r
				i2->view=3;
				break;
			case 38:	// a:q:q
				i2->filter_gb=!i2->filter_gb;
				break;
			case 39:	// o:s:s
				i2->filter_gr=!i2->filter_gr;
				break;
			case 52:	// ;:w:z
				i2->noise_f=!i2->noise_f;
				break;
			case 53:	// q:x:x
				i2->view=5;
				break;
			case 43:	// d:h:h
				i2->view=4;
				break;
			case 44:	// h:j:j
				film=1-film;
				break;
			case 45:	// t:k:k
				scrot();
				break;
			case 47:	// s:m:
				i2->flag=!i2->flag;
				break;
			case 100:	// left arrow
				i2->edge_factor--;
			printf("The new edge factor is %i\n",i2->edge_factor);
				break;
			case 102:	// right arrow
				i2->edge_factor++;
			printf("The new edge factor is %i\n",i2->edge_factor);
				break;
			default:
				printf("Key %i:%i pressed at (%i,%i)\n",
	event.xkey.state,event.xkey.keycode,event.xkey.x,event.xkey.y);
			}
		}; break;
		case KeyRelease:
		{
		}; break;
		
		default:
			break;
	}
      }
               
   }
   return 0;
}
